from fntypes.result import Error, Ok, Result

from mubble.api.api import API
from mubble.bot.dispatch.context import Context
from mubble.model import Model
from mubble.tools.adapter.abc import ABCAdapter
from mubble.tools.adapter.errors import AdapterError
from mubble.types.objects import Update


class RawEventAdapter(ABCAdapter[Update, Model]):
    def __init__(self, event_model: type[Model], /) -> None:
        self.event_model = event_model

    def __repr__(self) -> str:
        return "<{}: adapt Update -> {}>".format(
            self.__class__.__name__,
            self.event_model.__name__,
        )

    def adapt(self, api: API, update: Update, context: Context) -> Result[Model, AdapterError]:
        if isinstance(update.incoming_update, self.event_model):
            return Ok(update.incoming_update)
        return Error(AdapterError(f"Update is not an {self.event_model.__name__!r}."))


__all__ = ("RawEventAdapter",)



import typing_extensions as typing

from mubble.api.api import API, HTTPClient
from mubble.bot.dispatch import dispatch as dp
from mubble.bot.dispatch.abc import ABCDispatch
from mubble.bot.polling import polling as pg
from mubble.bot.polling.abc import ABCPolling
from mubble.modules import logger
from mubble.tools.loop_wrapper import ABCLoopWrapper
from mubble.tools.loop_wrapper import loop_wrapper as lw

Dispatch = typing.TypeVar(
    "Dispatch", bound=ABCDispatch, default=dp.Dispatch[HTTPClient]
)
Polling = typing.TypeVar("Polling", bound=ABCPolling, default=pg.Polling[HTTPClient])
LoopWrapper = typing.TypeVar(
    "LoopWrapper", bound=ABCLoopWrapper, default=lw.LoopWrapper
)


class Mubble(typing.Generic[HTTPClient, Dispatch, Polling, LoopWrapper]):
    def __init__(
        self,
        api: API[HTTPClient],
        *,
        dispatch: Dispatch | None = None,
        polling: Polling | None = None,
        loop_wrapper: LoopWrapper | None = None,
    ) -> None:
        self.api = api
        self.dispatch = typing.cast(Dispatch, dispatch or dp.Dispatch())
        self.polling = typing.cast(Polling, polling or pg.Polling(api))
        self.loop_wrapper = typing.cast(LoopWrapper, loop_wrapper or lw.LoopWrapper())

    def __repr__(self) -> str:
        return "<{}: api={!r}, dispatch={!r}, polling={!r}, loop_wrapper={!r}>".format(
            self.__class__.__name__,
            self.api,
            self.dispatch,
            self.polling,
            self.loop_wrapper,
        )

    @property
    def on(self) -> Dispatch:
        return self.dispatch

    async def reset_webhook(self) -> None:
        if not (await self.api.get_webhook_info()).unwrap().url:
            return
        await self.api.delete_webhook()

    async def run_polling(
        self,
        *,
        offset: int = 0,
        skip_updates: bool = False,
    ) -> typing.NoReturn:
        async def polling() -> typing.NoReturn:
            if skip_updates:
                logger.debug("Dropping pending updates")
                await self.reset_webhook()
                await self.api.delete_webhook(drop_pending_updates=True)
            self.polling.offset = offset

            async for updates in self.polling.listen():
                for update in updates:
                    logger.debug(
                        "Received update (update_id={}, update_type={!r})",
                        update.update_id,
                        update.update_type.name,
                    )
                    self.loop_wrapper.add_task(self.dispatch.feed(update, self.api))

        if self.loop_wrapper.is_running:
            await polling()
        else:
            self.loop_wrapper.add_task(polling())
            self.loop_wrapper.run_event_loop()

    def run_forever(
        self, *, offset: int = 0, skip_updates: bool = False
    ) -> typing.NoReturn:
        logger.debug("Running blocking polling (id={})", self.api.id)
        self.loop_wrapper.add_task(
            self.run_polling(offset=offset, skip_updates=skip_updates)
        )
        self.loop_wrapper.run_event_loop()


__all__ = ("Mubble",)



from __future__ import annotations

from abc import ABC

import typing_extensions as typing
from fntypes import Some

from mubble.api import API
from mubble.bot.cute_types.base import BaseCute
from mubble.bot.dispatch.context import Context
from mubble.model import Model
from mubble.modules import logger
from mubble.tools.adapter.abc import ABCAdapter, run_adapter
from mubble.tools.lifespan import Lifespan
from mubble.types.objects import Update

ToEvent = typing.TypeVar("ToEvent", bound=Model, default=typing.Any)


async def run_middleware[Event: Model, R: bool | None](
    method: typing.Callable[typing.Concatenate[Event, Context, ...], typing.Awaitable[R]],
    api: API[typing.Any],
    event: Event,
    ctx: Context,
    raw_event: Update | None = None,
    adapter: "ABCAdapter[Update, Event] | None" = None,
    *args: typing.Any,
    **kwargs: typing.Any,
) -> R:
    if adapter is not None:
        if raw_event is None:
            raise RuntimeError("raw_event must be specified to apply adapter")
        match await run_adapter(adapter, api, raw_event, ctx):
            case Some(val):
                event = val
            case _:
                return False  # type: ignore

    logger.debug("Running {}-middleware {!r}...", method.__name__, method.__qualname__.split(".")[0])
    return await method(event, ctx, *args, **kwargs)  # type: ignore


class ABCMiddleware[Event: Model | BaseCute](ABC):
    adapter: ABCAdapter[Update, Event] | None = None

    def __repr__(self) -> str:
        name = f"middleware {self.__class__.__name__!r}:"
        has_pre = self.pre.__qualname__.split(".")[0] != "ABCMiddleware"
        has_post = self.post.__qualname__.split(".")[0] != "ABCMiddleware"

        if has_post:
            name = "post-" + name
        if has_pre:
            name = "pre-" + name

        return "<{} with adapter={!r}>".format(name, self.adapter)

    async def pre(self, event: Event, ctx: Context) -> bool: ...

    async def post(self, event: Event, ctx: Context) -> None: ...

    @typing.overload
    def to_lifespan(self, event: Event, ctx: Context | None = None, *, api: API) -> Lifespan: ...

    @typing.overload
    def to_lifespan(self, event: Event, ctx: Context | None = None) -> Lifespan: ...

    def to_lifespan(
        self,
        event: Event,
        ctx: Context | None = None,
        api: API | None = None,
        **add_context: typing.Any,
    ) -> Lifespan:
        if api is None:
            if not isinstance(event, BaseCute):
                raise LookupError("Cannot get api, please pass as kwarg or provide BaseCute api-bound event")
            api = event.api

        ctx = ctx or Context()
        ctx |= add_context
        return Lifespan(
            startup_tasks=[run_middleware(self.pre, api, event, raw_event=None, ctx=ctx, adapter=None)],
            shutdown_tasks=[
                run_middleware(
                    self.post,
                    api,
                    event,
                    raw_event=None,
                    ctx=ctx,
                    adapter=None,
                )
            ],
        )


__all__ = ("ABCMiddleware", "run_middleware")



import abc
import dataclasses
import inspect
import typing

from fntypes import Error, Nothing, Ok, Option, Some
from fntypes.result import Result

from mubble.modules import logger
from mubble.tools.adapter.errors import AdapterError

type AdaptResult[To] = Result[To, AdapterError] | typing.Awaitable[Result[To, AdapterError]]


if typing.TYPE_CHECKING:
    from mubble.api.api import API
    from mubble.bot.dispatch.context import Context
    from mubble.model import Model


class ABCAdapter[From: "Model", To](abc.ABC):
    ADAPTED_VALUE_KEY: str | None = None

    @abc.abstractmethod
    def adapt(self, api: "API", update: From, context: "Context") -> AdaptResult[To]:
        pass


@dataclasses.dataclass(slots=True)
class Event[To]:
    obj: To


async def run_adapter[T, U: "Model"](
    adapter: "ABCAdapter[U, T]",
    api: "API",
    update: U,
    context: "Context",
) -> Option[T]:
    adapt_result = adapter.adapt(api, update, context)
    match await adapt_result if inspect.isawaitable(adapt_result) else adapt_result:
        case Ok(value):
            return Some(value)
        case Error(err):
            logger.debug("Adapter {!r} failed with error message: {!r}", adapter, str(err))
            return Nothing()


__all__ = ("ABCAdapter", "AdaptResult", "Event", "run_adapter")



import inspect
import typing

from fntypes.result import Error, Ok

from mubble.api.api import API
from mubble.bot.cute_types.update import UpdateCute
from mubble.bot.dispatch.context import Context
from mubble.modules import logger
from mubble.node.base import ComposeError, Node, get_nodes
from mubble.node.composer import CONTEXT_STORE_NODES_KEY, NodeSession, compose_nodes
from mubble.node.scope import NodeScope
from mubble.tools.magic import get_polymorphic_implementations, impl, magic_bundle
from mubble.types.objects import Update


class Polymorphic(Node):
    @classmethod
    async def compose(cls, raw_update: Update, update: UpdateCute, context: Context) -> typing.Any:
        logger.debug(f"Composing polymorphic node {cls.__name__!r}...")
        scope = getattr(cls, "scope", None)
        node_ctx = context.get_or_set(CONTEXT_STORE_NODES_KEY, {})
        data = {
            API: update.ctx_api,
            Context: context,
            Update: raw_update,
        }

        for i, impl_ in enumerate(get_polymorphic_implementations(cls)):
            logger.debug("Checking impl {!r}...", impl_.__name__)
            node_collection = None

            match await compose_nodes(get_nodes(impl_), context, data=data):
                case Ok(col):
                    node_collection = col
                case Error(err):
                    logger.debug(f"Composition failed with error: {err!r}")

            if node_collection is None:
                logger.debug("Impl {!r} composition failed!", impl_.__name__)
                continue

            # To determine whether this is a right morph, all subnodes must be resolved
            if scope is NodeScope.PER_EVENT and (cls, i) in node_ctx:
                logger.debug(
                    "Morph is already cached as per_event node, using its value. Impl {!r} succeeded!",
                    impl_.__name__,
                )
                res: NodeSession = node_ctx[(cls, i)]
                await node_collection.close_all()
                return res.value

            result = impl_(cls, **node_collection.values | magic_bundle(impl_, data, typebundle=True))
            if inspect.isawaitable(result):
                result = await result

            if scope is NodeScope.PER_EVENT:
                node_ctx[(cls, i)] = NodeSession(cls, result, {})

            await node_collection.close_all(with_value=result)
            logger.debug("Impl {!r} succeeded with value: {!r}", impl_.__name__, result)
            return result

        raise ComposeError("No implementation found.")


__all__ = ("Polymorphic", "impl")



import typing

import msgspec

from mubble.api.api import API
from mubble.model import Model, is_none


def compose_method_params[
    Cute: BaseCute
](
    params: dict[str, typing.Any],
    update: Cute,
    *,
    default_params: set[str | tuple[str, str]] | None = None,
    validators: dict[str, typing.Callable[[Cute], bool]] | None = None,
) -> dict[str, typing.Any]:
    default_params = default_params or set()
    validators = validators or {}

    for param in default_params:
        param_name = param if isinstance(param, str) else param[0]
        if param_name not in params:
            if param_name in validators and not validators[param_name](update):
                continue
            params[param_name] = getattr(
                update, param if isinstance(param, str) else param[1]
            )

    return params


if typing.TYPE_CHECKING:
    from mubble.node.base import Node

    class BaseCute[Update: Model](Model):
        api: API

        @classmethod
        def as_node(cls) -> type[Node]: ...

        @classmethod
        def from_update(cls, update: Update, bound_api: API) -> typing.Self: ...

        @property
        def ctx_api(self) -> API: ...

        def to_dict(
            self,
            *,
            exclude_fields: set[str] | None = None,
        ) -> dict[str, typing.Any]:
            """:param exclude_fields: Cute model field names to exclude from the dictionary representation of this cute model.
            :return: A dictionary representation of this cute model.
            """
            ...

        def to_full_dict(
            self,
            *,
            exclude_fields: set[str] | None = None,
        ) -> dict[str, typing.Any]:
            """:param exclude_fields: Cute model field names to exclude from the dictionary representation of this cute model.
            :return: A dictionary representation of this model including all models, structs, custom types.
            """
            ...

else:
    import msgspec
    from fntypes.co import Nothing, Some, Variative

    from mubble.msgspec_utils import Option, decoder, encoder, struct_as_dict
    from mubble.msgspec_utils import get_class_annotations as _get_class_annotations

    def _get_cute_from_generic(generic_args, /):
        for arg in generic_args:
            orig_arg = typing.get_origin(arg) or arg

            if not isinstance(orig_arg, type):
                continue
            if orig_arg in (Variative, Some, Option):
                return _get_cute_from_generic(typing.get_args(arg))
            if issubclass(orig_arg, BaseCute):
                return arg

        return None

    def _get_cute_annotations(annotations, /):
        cute_annotations = {}

        for key, hint in annotations.items():
            if not isinstance(hint, type):
                if (cute := _get_cute_from_generic(typing.get_args(hint))) is not None:
                    cute_annotations[key] = cute

            elif issubclass(hint, BaseCute):
                cute_annotations[key] = hint

        return cute_annotations

    def _validate_value(value, /):
        while isinstance(value, Variative | Some):
            if isinstance(value, Variative):
                value = value.v
            if isinstance(value, Some):
                value = value.value
        return value

    class BaseCute[Update]:
        api: API

        def __init_subclass__(cls, *args, **kwargs):
            super().__init_subclass__(*args, **kwargs)

            if not cls.__bases__ or not issubclass(cls.__bases__[0], BaseCute):
                return

            cls.__is_solved_annotations__ = False
            cls.__cute_annotations__ = None
            cls.__event_node__ = None

        @classmethod
        def as_node(cls):
            if cls.__event_node__ is None:
                from mubble.node.event import _EventNode

                cls.__event_node__ = _EventNode[cls]
            return cls.__event_node__

        @classmethod
        def from_update(cls, update, bound_api):
            if not cls.__is_solved_annotations__:
                cls.__is_solved_annotations__ = True
                cls.__annotations__ = _get_class_annotations(cls)

            if cls.__cute_annotations__ is None:
                cls.__cute_annotations__ = _get_cute_annotations(cls.__annotations__)

            return cls(
                **{
                    field: (
                        decoder.convert(
                            cls.__cute_annotations__[field].from_update(
                                _validate_value(value), bound_api=bound_api
                            ),
                            type=cls.__annotations__[field],
                        )
                        if field in cls.__cute_annotations__
                        and not isinstance(value, Nothing | msgspec.UnsetType)
                        else value
                    )
                    for field, value in update.to_dict().items()
                },
                api=bound_api,
            )

        @property
        def ctx_api(self):
            return self.api

        def _to_dict(self, dct_name, exclude_fields, full):
            if dct_name not in self.__dict__:
                self.__dict__[dct_name] = (
                    struct_as_dict(self)
                    if not full
                    else encoder.to_builtins(
                        {
                            k: (
                                field.to_dict(exclude_fields=exclude_fields)
                                if isinstance(field, BaseCute)
                                else field
                            )
                            for k in self.__struct_fields__
                            if k not in exclude_fields
                            and not isinstance(
                                field := _validate_value(getattr(self, k)),
                                msgspec.UnsetType,
                            )
                            and not is_none(field)
                        },
                        order="deterministic",
                    )
                )

            if not exclude_fields:
                return self.__dict__[dct_name]

            return {
                key: value
                for key, value in self.__dict__[dct_name].items()
                if key not in exclude_fields
            }

        def to_dict(self, *, exclude_fields=None, full=False):
            exclude_fields = exclude_fields or set()
            return self._to_dict(
                "model_as_dict", exclude_fields={"api"} | exclude_fields, full=full
            )

        def to_full_dict(self, *, exclude_fields=None):
            return self.to_dict(exclude_fields=exclude_fields, full=True)


__all__ = ("BaseCute", "compose_method_params")



from .abc import ABCClient
from .aiohttp import AiohttpClient
from .form_data import MultipartFormProto, encode_form_data
from .sonic import AiosonicClient

__all__ = (
    "ABCClient",
    "AiohttpClient",
    "AiosonicClient",
    "MultipartFormProto",
    "encode_form_data",
)



from functools import cached_property

import msgspec
import typing_extensions as typing
from fntypes.result import Error, Ok, Result

from mubble.api.error import APIError
from mubble.api.response import APIResponse
from mubble.api.token import Token
from mubble.client import ABCClient, AiohttpClient, MultipartFormProto
from mubble.model import decoder
from mubble.types.methods import APIMethods

HTTPClient = typing.TypeVar("HTTPClient", bound=ABCClient, default=AiohttpClient)

type Json = str | int | float | bool | list[Json] | dict[str, Json] | None


def compose_data[MultipartForm: MultipartFormProto](
    client: ABCClient[MultipartForm],
    data: dict[str, typing.Any],
    files: dict[str, tuple[str, bytes]],
) -> MultipartForm:
    if not data and not files:
        return client.multipart_form_factory()
    return client.get_form(data=data, files=files)


class API(APIMethods[HTTPClient], typing.Generic[HTTPClient]):
    """Bot API with available API methods and http client."""

    API_URL = "https://api.telegram.org/"
    API_FILE_URL = "https://api.telegram.org/file/"

    token: Token
    http: HTTPClient

    def __init__(self, token: Token, *, http: HTTPClient | None = None) -> None:
        self.token = token
        self.http = http or AiohttpClient()  # type: ignore
        super().__init__(api=self)

    def __repr__(self) -> str:
        return "<{}: token={!r}, http={!r}>".format(
            self.__class__.__name__,
            self.token,
            self.http,
        )

    @cached_property
    def id(self) -> int:
        return self.token.bot_id

    @property
    def request_url(self) -> str:
        return self.API_URL + f"bot{self.token}/"

    @property
    def request_file_url(self) -> str:
        return self.API_FILE_URL + f"bot{self.token}/"

    async def download_file(self, file_path: str) -> bytes:
        return await self.http.request_content(f"{self.request_file_url}/{file_path}")

    async def request(
        self,
        method: str,
        data: dict[str, typing.Any] | None = None,
        files: dict[str, tuple[str, bytes]] | None = None,
    ) -> Result[Json, APIError]:
        """Request a `JSON` response with the `POST` HTTP method and passing data, files as `multipart/form-data`."""
        response = await self.http.request_json(
            url=self.request_url + method,
            method="POST",
            data=compose_data(self.http, data or {}, files or {}),
        )
        if response.get("ok", False) is True:
            return Ok(response["result"])
        return Error(
            APIError(
                code=response.get("error_code", 400),
                error=response.get("description", "Something went wrong"),
            ),
        )

    async def request_raw(
        self,
        method: str,
        data: dict[str, typing.Any] | None = None,
        files: dict[str, tuple[str, bytes]] | None = None,
    ) -> Result[msgspec.Raw, APIError]:
        """Request a `raw` response with the `POST` HTTP method and passing data, files as `multipart/form-data`."""
        response_bytes = await self.http.request_bytes(
            url=self.request_url + method,
            method="POST",
            data=compose_data(self.http, data or {}, files or {}),
        )
        return decoder.decode(response_bytes, type=APIResponse).to_result()


__all__ = ("API",)



from mubble_old import Token, API, Mubble
from mubble_old.modules import logger

api = API(Token.from_env())
bot = Mubble(api)
logger.set_level("INFO")


@bot.loop_wrapper.timer(seconds=5)
async def once():
    print("I only fire once after startup, and that's it.")


@bot.loop_wrapper.interval(minutes=1)
async def repeat():
    print("I kick in every minute.")


@bot.loop_wrapper.lifespan.on_startup
async def startup():
    print("I'm working on startup.")


@bot.loop_wrapper.lifespan.on_shutdown
async def shutdown():
    print("I'm working on shutdown.")


bot.run_forever()



from mubble_old import InlineButton, InlineKeyboard, Token, API, Mubble, Message
from mubble_old.bot.dispatch.context import Context
from mubble_old.bot.rules.callback_data import CallbackQuery
from mubble_old.rules import MessageRule, CallbackQueryRule, Command, CallbackData
from mubble_old.modules import logger

api = API(Token.from_env())
bot = Mubble(api)

logger.set_level("INFO")
OWNER_ID = 1734816882  # Use your Telegram id

keyboard = (
    InlineKeyboard().add(InlineButton("👋🏼 Bye!", callback_data="bye"))
).get_markup()


class HasPhoto(MessageRule):
    async def check(self, message: Message, ctx: Context) -> bool:
        return bool(message.photo and message.photo.unwrap())


class IsOwnerMessage(MessageRule):
    async def check(self, message: Message, ctx: Context) -> bool:
        return message.from_user.id == OWNER_ID


class IsOwnerCallback(CallbackQueryRule):
    async def check(self, event: CallbackQuery, ctx: Context) -> bool:
        return event.from_user.id == OWNER_ID


@bot.on.message(HasPhoto())
async def photo_handler(message: Message):
    await message.answer("📸 Looks like you just dropped a pic.")


@bot.on.message(Command("hello") & IsOwnerMessage())
async def owner_hello_handler(message: Message):
    await message.answer("🥳 Hi, owner!", reply_markup=keyboard)


@bot.on.message(Command("hello"))
async def hello_handler(message: Message):
    await message.answer("🥳 Hi, user!", reply_markup=keyboard)


@bot.on.callback_query(CallbackData("bye") & IsOwnerCallback())
async def owner_bye_handler(cq: CallbackQuery):
    await cq.edit_text("👋🏼 Bye, owner!")


@bot.on.callback_query(CallbackData("bye"))
async def bye_handler(cq: CallbackQuery):
    await cq.edit_text("👋🏼 Bye, user!")


bot.run_forever()



import pathlib

from mubble_old import Token, API, Mubble, Message
from mubble_old.modules import logger
from mubble_old.rules import Command
from mubble_old.types import InputFile

api = API(token=Token.from_env())
bot = Mubble(api)
logger.set_level("INFO")

mubble_logo = pathlib.Path("images/mubble_logo.png").read_bytes()


@bot.on.message(Command("photo"))
async def start(message: Message):
    await message.answer_photo(
        InputFile("mubble.png", mubble_logo),
        caption="📸 Here's a pic for you!",
    )


bot.run_forever()



from .adapter import (
    ABCAdapter,
    DataclassAdapter,
    EventAdapter,
    NodeAdapter,
    RawEventAdapter,
    RawUpdateAdapter,
)
from .buttons import BaseButton
from .callback_data_serilization import (
    ABCDataSerializer,
    JSONSerializer,
    MsgPackSerializer,
)
from .error_handler import ABCErrorHandler, Catcher, CatcherError, ErrorHandler
from .formatting import (
    Base,
    BlockQuote,
    FormatString,
    HTMLFormatter,
    Link,
    Mention,
    PreCode,
    SpecialFormat,
    TgEmoji,
    block_quote,
    bold,
    code_inline,
    escape,
    italic,
    link,
    mention,
    pre_code,
    spoiler,
    strike,
    tg_bot_attach_open_any_chat,
    tg_bot_attach_open_current_chat,
    tg_bot_attach_open_specific_chat,
    tg_bot_start_link,
    tg_bot_startchannel_link,
    tg_bot_startgroup_link,
    tg_chat_folder_link,
    tg_chat_invite_link,
    tg_direct_mini_app_link,
    tg_emoji,
    tg_emoji_link,
    tg_emoji_stickerset_link,
    tg_invoice_link,
    tg_language_pack_link,
    tg_main_mini_app_link,
    tg_mention_link,
    tg_open_message_link,
    tg_premium_multigift_link,
    tg_premium_offer_link,
    tg_private_channel_boost_link,
    tg_private_message_link,
    tg_public_channel_boost_link,
    tg_public_message_link,
    tg_public_username_link,
    tg_share_link,
    tg_story_link,
    underline,
)
from .functional import from_optional
from .global_context import (
    ABCGlobalContext,
    CtxVar,
    GlobalContext,
    GlobalCtxVar,
    MubbleContext,
    ctx_var,
)
from .i18n import (
    ABCI18n,
    ABCTranslator,
    ABCTranslatorMiddleware,
    I18nEnum,
    SimpleI18n,
    SimpleTranslator,
)
from .input_file_directory import InputFileDirectory
from .keyboard import (
    AnyMarkup,
    Button,
    InlineButton,
    InlineKeyboard,
    Keyboard,
    RowButtons,
)
from .limited_dict import LimitedDict
from .loop_wrapper import ABCLoopWrapper, DelayedTask, Lifespan, LoopWrapper
from .magic import (
    cancel_future,
    get_annotations,
    get_cached_translation,
    get_default_args,
    get_func_parameters,
    get_polymorphic_implementations,
    impl,
    magic_bundle,
    resolve_arg_names,
)
from .parse_mode import ParseMode
from .state_storage import ABCStateStorage, MemoryStateStorage, StateData

__all__ = (
    "ABCAdapter",
    "ABCDataSerializer",
    "ABCErrorHandler",
    "ABCGlobalContext",
    "ABCI18n",
    "ABCLoopWrapper",
    "ABCStateStorage",
    "ABCTranslator",
    "ABCTranslatorMiddleware",
    "AnyMarkup",
    "Base",
    "BaseButton",
    "BlockQuote",
    "Button",
    "Catcher",
    "CatcherError",
    "CtxVar",
    "DataclassAdapter",
    "DelayedTask",
    "ErrorHandler",
    "EventAdapter",
    "FormatString",
    "GlobalContext",
    "GlobalCtxVar",
    "HTMLFormatter",
    "I18nEnum",
    "InlineButton",
    "InlineKeyboard",
    "InputFileDirectory",
    "JSONSerializer",
    "Keyboard",
    "Lifespan",
    "LimitedDict",
    "Link",
    "LoopWrapper",
    "MemoryStateStorage",
    "Mention",
    "MsgPackSerializer",
    "NodeAdapter",
    "ParseMode",
    "PreCode",
    "RawEventAdapter",
    "RawUpdateAdapter",
    "RowButtons",
    "SimpleI18n",
    "SimpleTranslator",
    "SpecialFormat",
    "StateData",
    "MubbleContext",
    "TgEmoji",
    "block_quote",
    "bold",
    "cancel_future",
    "code_inline",
    "ctx_var",
    "escape",
    "from_optional",
    "get_annotations",
    "get_cached_translation",
    "get_default_args",
    "get_func_parameters",
    "get_polymorphic_implementations",
    "impl",
    "italic",
    "link",
    "magic_bundle",
    "mention",
    "pre_code",
    "resolve_arg_names",
    "spoiler",
    "strike",
    "tg_bot_attach_open_any_chat",
    "tg_bot_attach_open_current_chat",
    "tg_bot_attach_open_specific_chat",
    "tg_bot_start_link",
    "tg_bot_startchannel_link",
    "tg_bot_startgroup_link",
    "tg_chat_folder_link",
    "tg_chat_invite_link",
    "tg_direct_mini_app_link",
    "tg_emoji",
    "tg_emoji_link",
    "tg_emoji_stickerset_link",
    "tg_invoice_link",
    "tg_language_pack_link",
    "tg_main_mini_app_link",
    "tg_mention_link",
    "tg_open_message_link",
    "tg_premium_multigift_link",
    "tg_premium_offer_link",
    "tg_private_channel_boost_link",
    "tg_private_message_link",
    "tg_public_channel_boost_link",
    "tg_public_message_link",
    "tg_public_username_link",
    "tg_share_link",
    "tg_story_link",
    "underline",
)



from mubble_old import Token, API, Mubble, WaiterMachine, Message, Choice
from mubble_old.rules import Command
from mubble_old.modules import logger

api = API(token=Token.from_env())
bot = Mubble(api)
wm = WaiterMachine()

logger.set_level("INFO")


devices = {
    "iphone": ["📱 iPhone 15 Pro", 999.0],
    "ipad": ["📲 iPad Pro M4", 1299.0],
    "apple_watch": ["⌚️ Apple Watch Ultra 2", 799.0],
    "macbook": ["💻 Macbook Pro 16 M3 Pro", 2499.0],
}


@bot.on.message(Command("choice"))
async def action(message: Message):
    chosen, message_id = await (
        Choice(
            wm,
            message.chat.id,
            "Pick one of these devices.",
            ready_text="Pick!",
            max_in_row=2,
        )
        .add_option("iphone", "iPhone", "iPhone 🟢", is_picked=True)
        .add_option("ipad", "iPad", "iPad 🟢")
        .add_option("apple_watch", "Apple watch", "Apple watch 🟢")
        .add_option("macbook", "Macbook", "Macbook 🟢")
        .wait(message.ctx_api, bot.dispatch.callback_query)
    )

    name, price = devices[chosen]
    await message.edit(
        text=f"You've picked the {name}, which goes for {price:.2f} bucks.",
        message_id=message_id,
    )


bot.run_forever()



import typing

from mubble_old import API, Message, Mubble, ParseMode, Token
from mubble_old.modules import logger
from mubble_old.rules import Markup

api = API(token=Token.from_env())
bot = Mubble(api)
logger.set_level("INFO")


@bot.on.message(
    Markup(
        [
            "/solve <a:int> <(+-/*)^operator> <b:int>",
            "/solve <a:int><(+-/*)^operator><b:int>",
        ]
    )
)
async def solve(
    message: Message,
    a: int,
    b: int,
    operator: typing.Literal["+", "-", "/", "*"],
):
    statement = f"{a} {operator} {b}"
    await message.reply(
        f"🤓 Result: <code>{statement} = {eval(statement):.2f}</code>",
        parse_mode=ParseMode.HTML,
    )


@solve.error_handler(ZeroDivisionError)
async def zero_division_catcher(_: ZeroDivisionError) -> str:
    return f"🤓 Error: you can't divide by zero! {_}"


bot.run_forever()



class APIError(Exception):
    def __init__(self, code: int, error: str) -> None:
        self.code, self.error = code, error

    def __str__(self) -> str:
        return f"[{self.code}] {self.error}"

    def __repr__(self) -> str:
        return f"<APIError: {self.__str__()}>"


class APIServerError(Exception):
    pass


class InvalidTokenError(BaseException):
    pass


__all__ = ("APIError", "APIServerError", "InvalidTokenError")



from fntypes.option import Nothing, Some
from fntypes.result import Error, Ok, Result

from mubble.api.api import API
from mubble.bot.cute_types.base import BaseCute
from mubble.bot.cute_types.update import UpdateCute
from mubble.bot.dispatch.context import Context
from mubble.tools.adapter.abc import ABCAdapter
from mubble.tools.adapter.errors import AdapterError
from mubble.types.enums import UpdateType
from mubble.types.objects import Update


class DataclassAdapter[Dataclass](ABCAdapter[Update, Dataclass]):
    ADAPTED_VALUE_KEY: str

    def __init__(
        self,
        dataclass: type[Dataclass],
        update_type: UpdateType | None = None,
    ) -> None:
        self.ADAPTED_VALUE_KEY = f"_adapted_dataclass_{dataclass.__name__}"
        self.dataclass = dataclass
        self.update_type = update_type

    def __repr__(self) -> str:
        return f"<Update -> {self.dataclass.__name__}>"

    def adapt(self, api: API, update: Update, context: Context) -> Result[Dataclass, AdapterError]:
        if self.ADAPTED_VALUE_KEY in context:
            return Ok(context[self.ADAPTED_VALUE_KEY])

        update_type = (self.update_type or update.update_type).value
        try:
            if self.dataclass is Update:
                return Ok(update)  # type: ignore
            elif issubclass(self.dataclass, UpdateCute):
                dataclass = self.dataclass.from_update(update, bound_api=api)
            else:
                match getattr(update, update_type):
                    case Some(val):
                        dataclass = (
                            self.dataclass.from_update(val, bound_api=api)
                            if issubclass(self.dataclass, BaseCute)
                            else self.dataclass(**val.to_dict())
                        )
                    case Nothing():
                        return Error(AdapterError(f"Update has no event {update_type!r}."))
        except Exception as e:
            return Error(AdapterError(f"Cannot adapt Update to {self.dataclass!r}, error: {e!r}"))

        context[self.ADAPTED_VALUE_KEY] = dataclass
        return Ok(dataclass)  # type: ignore


__all__ = ("DataclassAdapter",)



from .attachment import (
    Animation,
    Attachment,
    Audio,
    Document,
    Photo,
    SuccessfulPayment,
    Video,
    VideoNote,
    Voice,
)
from .base import (
    Composable,
    ComposeError,
    DataNode,
    FactoryNode,
    GlobalNode,
    IsNode,
    Name,
    Node,
    NodeComposeFunction,
    NodeImpersonation,
    NodeProto,
    NodeType,
    as_node,
    is_node,
    scalar_node,
    unwrap_node,
)
from .callback_query import (
    CallbackQueryData,
    CallbackQueryDataJson,
    Field,
)
from .command import CommandInfo
from .composer import NodeCollection, NodeSession, compose_node, compose_nodes
from .container import ContainerNode
from .either import Either, Optional
from .event import EventNode
from .file import File, FileId
from .me import Me
from .payload import Payload, PayloadData, PayloadSerializer
from .polymorphic import Polymorphic, impl
from .rule import RuleChain
from .scope import (
    GLOBAL,
    PER_CALL,
    PER_EVENT,
    NodeScope,
    global_node,
    per_call,
    per_event,
)
from .source import ChatSource, Source, UserId, UserSource
from .text import Text, TextInteger, TextLiteral
from .tools import generate_node

__all__ = (
    "Animation",
    "Attachment",
    "Audio",
    "CallbackQueryData",
    "CallbackQueryDataJson",
    "ChatSource",
    "CommandInfo",
    "Composable",
    "ComposeError",
    "ContainerNode",
    "DataNode",
    "Document",
    "Either",
    "EventNode",
    "FactoryNode",
    "Field",
    "Field",
    "File",
    "FileId",
    "GLOBAL",
    "GlobalNode",
    "IsNode",
    "Me",
    "Name",
    "Node",
    "NodeCollection",
    "NodeComposeFunction",
    "NodeImpersonation",
    "NodeProto",
    "NodeScope",
    "NodeSession",
    "NodeType",
    "Optional",
    "PER_CALL",
    "PER_EVENT",
    "Payload",
    "PayloadData",
    "PayloadSerializer",
    "Photo",
    "Polymorphic",
    "RuleChain",
    "Source",
    "SuccessfulPayment",
    "Text",
    "TextInteger",
    "TextLiteral",
    "UserId",
    "UserSource",
    "Video",
    "VideoNote",
    "Voice",
    "as_node",
    "compose_node",
    "compose_nodes",
    "generate_node",
    "global_node",
    "impl",
    "is_node",
    "per_call",
    "per_event",
    "scalar_node",
    "unwrap_node",
)



from mubble_old import (
    API,
    ABCMiddleware,
    CallbackQuery,
    InlineButton,
    InlineKeyboard,
    Message,
    Mubble,
    ParseMode,
    Token,
)
from mubble_old.bot import Context
from mubble_old.bot.rules.callback_data import CallbackData
from mubble_old.modules import logger
from mubble_old.rules import Command
from mubble_old.tools import GlobalContext

api = API(token=Token.from_env())
bot = Mubble(api)
global_ctx = GlobalContext(msg_counter=dict(), cb_counter=dict())

logger.set_level("INFO")

keyboard = (
    InlineKeyboard()
    .add(InlineButton("🍪 Click", callback_data="click"))
    .row()
    .add(InlineButton("👉🏼 Clicks count", callback_data="count"))
).get_markup()


@bot.on.message.register_middleware()
class MessageContextMiddleware(ABCMiddleware[Message]):
    async def pre(self, event: Message, ctx: Context) -> bool:
        counter = global_ctx.get_value("msg_counter", dict[int, int]).unwrap()
        counter[event.from_user.id] = counter.get(event.from_user.id, 0) + 1

        ctx.set("msg_count", counter[event.from_user.id])
        return True


@bot.on.callback_query.register_middleware()
class CallbackContextMiddleware(ABCMiddleware[CallbackQuery]):
    async def pre(self, event: CallbackQuery, ctx: Context) -> bool:
        counter = global_ctx.get_value("cb_counter", dict[int, int]).unwrap()
        counter[event.from_user.id] = counter.get(event.from_user.id, 0) + 1

        ctx.set("cb_count", counter[event.from_user.id])
        return True


@bot.on.message(Command("count"))
async def count_handler(message: Message, msg_count: int):
    await message.answer(
        f"💬 <a href='tg://user?id={message.from_user.id}'>{message.from_user.first_name}</a>, you've sent {msg_count} messages!",
        parse_mode=ParseMode.HTML,
        reply_markup=keyboard,
    )


@bot.on.callback_query(CallbackData("click"))
async def click_handler(cq: CallbackQuery, cb_count: int):
    await cq.answer(f"Total: {cb_count} clicks")


@bot.on.callback_query(CallbackData("count"))
async def clicks_count_handler(cq: CallbackQuery, cb_count: int):
    await cq.answer()
    await cq.edit_text(
        f"💬 <a href='tg://user?id={cq.from_user.id}'>{cq.from_user.first_name}</a>, you've clicked the buttons {cb_count} times!",
        parse_mode=ParseMode.HTML,
    )


bot.run_forever()



import asyncio
import contextlib
import datetime
import typing

from mubble.modules import logger
from mubble.tools.lifespan import (
    CoroutineFunc,
    CoroutineTask,
    DelayedTask,
    Lifespan,
    Task,
    to_coroutine_task,
)
from mubble.tools.loop_wrapper.abc import ABCLoopWrapper
from mubble.tools.magic import cancel_future


class LoopWrapper(ABCLoopWrapper):
    def __init__(
        self,
        *,
        tasks: list[CoroutineTask[typing.Any]] | None = None,
        lifespan: Lifespan | None = None,
    ) -> None:
        self.tasks: list[CoroutineTask[typing.Any]] = tasks or []
        self.lifespan = lifespan or Lifespan()
        self._loop: asyncio.AbstractEventLoop | None = None

    @property
    def is_running(self) -> bool:
        if self._loop is None:
            return False
        return self._loop.is_running()

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        assert self._loop is not None, "Loop is not set."
        return self._loop

    def __repr__(self) -> str:
        return "<{}: loop={!r} with tasks={!r}, lifespan={!r}>".format(
            self.__class__.__name__,
            self._loop,
            self.tasks,
            self.lifespan,
        )

    async def _run_tasks(self) -> None:
        async with asyncio.TaskGroup() as tg:
            while self.tasks:
                tg.create_task(self.tasks.pop(0))

    def run_event_loop(self) -> typing.NoReturn:  # type: ignore
        if not self.tasks:
            logger.warning("Run loop without tasks!")

        try:
            self._loop = asyncio.get_running_loop()
        except RuntimeError:
            self._loop = asyncio.get_event_loop()

        self.lifespan.start()
        self._loop.create_task(self._run_tasks())

        tasks = asyncio.all_tasks(self._loop)
        try:
            while tasks:
                tasks_results, _ = self._loop.run_until_complete(
                    asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION),
                )
                for task_result in tasks_results:
                    try:
                        task_result.result()
                    except BaseException:
                        logger.exception("Traceback message below:")
                tasks = asyncio.all_tasks(self._loop)
        except KeyboardInterrupt:
            print()  # blank print for ^C
            logger.info("Caught KeyboardInterrupt, cancellation...")
            self.complete_tasks(tasks)
        finally:
            self.lifespan.shutdown()
            if self._loop.is_running():
                self._loop.close()

    def add_task(self, task: Task[..., typing.Any], /) -> None:
        task = to_coroutine_task(task)

        if self._loop is not None and self._loop.is_running():
            self._loop.create_task(task)
        else:
            self.tasks.append(task)

    def complete_tasks(self, tasks: set[asyncio.Task[typing.Any]], /) -> None:
        tasks = tasks | asyncio.all_tasks(self.loop)
        with contextlib.suppress(asyncio.CancelledError, asyncio.InvalidStateError):
            self.loop.run_until_complete(cancel_future(asyncio.gather(*tasks, return_exceptions=True)))

    @typing.overload
    def timer(self, *, seconds: datetime.timedelta) -> typing.Callable[..., typing.Any]: ...

    @typing.overload
    def timer(
        self,
        *,
        days: int = 0,
        hours: int = 0,
        minutes: int = 0,
        seconds: float = 0,
    ) -> typing.Callable[..., typing.Any]: ...

    def timer(
        self,
        *,
        days: int = 0,
        hours: int = 0,
        minutes: int = 0,
        seconds: float | datetime.timedelta = 0,
    ) -> typing.Callable[..., typing.Any]:
        if isinstance(seconds, datetime.timedelta):
            seconds = seconds.total_seconds()

        seconds += minutes * 60
        seconds += hours * 60 * 60
        seconds += days * 24 * 60 * 60

        def decorator[Func: CoroutineFunc[..., typing.Any]](func: Func) -> Func:
            self.add_task(DelayedTask(func, seconds, repeat=False))
            return func

        return decorator

    @typing.overload
    def interval(self, *, seconds: datetime.timedelta) -> typing.Callable[..., typing.Any]: ...

    @typing.overload
    def interval(
        self,
        *,
        days: int = 0,
        hours: int = 0,
        minutes: int = 0,
        seconds: float = 0,
    ) -> typing.Callable[..., typing.Any]: ...

    def interval(
        self,
        *,
        days: int = 0,
        hours: int = 0,
        minutes: int = 0,
        seconds: float | datetime.timedelta = 0,
    ) -> typing.Callable[..., typing.Any]:
        if isinstance(seconds, datetime.timedelta):
            seconds = seconds.total_seconds()

        seconds += minutes * 60
        seconds += hours * 60 * 60
        seconds += days * 24 * 60 * 60

        def decorator[Func: CoroutineFunc[..., typing.Any]](func: Func) -> Func:
            self.add_task(DelayedTask(func, seconds, repeat=True))
            return func

        return decorator


__all__ = ("DelayedTask", "LoopWrapper", "to_coroutine_task")



import abc
import inspect
import typing
from contextlib import suppress

from mubble.bot.cute_types import CallbackQueryCute
from mubble.bot.dispatch.context import Context
from mubble.bot.rules.abc import ABCRule, CheckResult
from mubble.bot.rules.payload import (
    PayloadEqRule,
    PayloadJsonEqRule,
    PayloadMarkupRule,
    PayloadModelRule,
)
from mubble.tools.adapter.event import EventAdapter
from mubble.types.enums import UpdateType

CallbackQuery: typing.TypeAlias = CallbackQueryCute
Validator: typing.TypeAlias = typing.Callable[[typing.Any], bool | typing.Awaitable[bool]]
MapDict: typing.TypeAlias = dict[str, "typing.Any | type[typing.Any] | Validator | list[MapDict] | MapDict"]
CallbackMap: typing.TypeAlias = list[tuple[str, "typing.Any | type[typing.Any] | Validator | CallbackMap"]]
CallbackMapStrict: typing.TypeAlias = list[tuple[str, "Validator | CallbackMapStrict"]]
CallbackDataEq: typing.TypeAlias = PayloadEqRule
CallbackDataJsonEq: typing.TypeAlias = PayloadJsonEqRule
CallbackDataMarkup: typing.TypeAlias = PayloadMarkupRule
CallbackDataJsonModel: typing.TypeAlias = PayloadModelRule


class CallbackQueryRule(
    ABCRule[CallbackQuery],
    abc.ABC,
    adapter=EventAdapter(UpdateType.CALLBACK_QUERY, CallbackQuery),
):
    @abc.abstractmethod
    def check(self, *args: typing.Any, **kwargs: typing.Any) -> CheckResult: ...


class HasData(CallbackQueryRule):
    def check(self, event: CallbackQuery) -> bool:
        return bool(event.data)


class CallbackQueryDataRule(CallbackQueryRule, abc.ABC, requires=[HasData()]):
    pass


class CallbackDataMap(CallbackQueryDataRule):
    def __init__(self, mapping: MapDict, /) -> None:
        self.mapping = self.transform_to_callbacks(
            self.transform_to_map(mapping),
        )

    @classmethod
    def transform_to_map(cls, mapping: MapDict) -> CallbackMap:
        """Transforms MapDict to CallbackMap."""
        callback_map = []

        for k, v in mapping.items():
            if isinstance(v, dict):
                v = cls.transform_to_map(v)
            callback_map.append((k, v))

        return callback_map

    @classmethod
    def transform_to_callbacks(cls, callback_map: CallbackMap) -> CallbackMapStrict:
        """Transforms `CallbackMap` to `CallbackMapStrict`."""
        callback_map_result = []

        for key, value in callback_map:
            if isinstance(value, type):
                validator = (lambda tp: lambda v: isinstance(v, tp))(value)
            elif isinstance(value, list):
                validator = cls.transform_to_callbacks(value)
            elif not callable(value):
                validator = (lambda val: lambda v: val == v)(value)
            else:
                validator = value
            callback_map_result.append((key, validator))

        return callback_map_result

    @staticmethod
    async def run_validator(value: typing.Any, validator: Validator) -> bool:
        """Run async or sync validator."""
        with suppress(BaseException):
            result = validator(value)
            if inspect.isawaitable(result):
                result = await result
            return result

        return False

    @classmethod
    async def match(cls, callback_data: dict[str, typing.Any], callback_map: CallbackMapStrict) -> bool:
        """Matches callback_data with callback_map recursively."""
        for key, validator in callback_map:
            if key not in callback_data:
                return False

            if isinstance(validator, list):
                if not (isinstance(callback_data[key], dict) and await cls.match(callback_data[key], validator)):
                    return False

            elif not await cls.run_validator(callback_data[key], validator):
                return False

        return True

    async def check(self, event: CallbackQuery, ctx: Context) -> bool:
        callback_data = event.decode_data().unwrap_or_none()
        if callback_data is None:
            return False
        if await self.match(callback_data, self.mapping):
            ctx.update(callback_data)
            return True
        return False


__all__ = (
    "CallbackDataEq",
    "CallbackDataJsonEq",
    "CallbackDataJsonModel",
    "CallbackDataMap",
    "CallbackDataMarkup",
    "CallbackQueryDataRule",
    "CallbackQueryRule",
    "HasData",
)



from mubble_old import API, Checkbox, Message, Mubble, Token, WaiterMachine
from mubble_old.modules import logger
from mubble_old.rules import Command

api = API(token=Token.from_env())
bot = Mubble(api)
wm = WaiterMachine()

logger.set_level("INFO")


@bot.on.message(Command("checkbox"))
async def action(message: Message):
    picked, message_id = await (
        Checkbox(
            wm,
            message.chat.id,
            "Pick the fruits you vibe with!",
            ready_text="✅ Done!",
            max_in_row=2,
        )
        .add_option("apple", "🍏", "Apple")
        .add_option("kiwi", "🥝", "Kiwi", is_picked=True)
        .add_option("strawberry", "🍓", "Strawberry")
        .add_option("orange", "🍊", "Orange", is_picked=True)
        .add_option("peach", "🍑", "Peach")
        .wait(message.ctx_api, bot.dispatch.callback_query)
    )

    await message.edit(
        text=f"You picked: {", ".join(name for name in picked if picked[name])}",
        chat_id=message.chat.id,
        message_id=message_id,
    )


bot.run_forever()



from mubble.bot.cute_types.chat_join_request import ChatJoinRequestCute
from mubble.bot.dispatch.view.base import BaseStateView


class ChatJoinRequestView(BaseStateView[ChatJoinRequestCute]):
    @classmethod
    def get_state_key(cls, event: ChatJoinRequestCute) -> int | None:
        return event.chat_id


__all__ = ("ChatJoinRequestView",)



from __future__ import annotations

import dataclasses

import typing_extensions as typing
from fntypes import Nothing, Option, Some
from vbml.patcher import Patcher

from mubble.api.api import API, HTTPClient
from mubble.bot.dispatch.abc import ABCDispatch
from mubble.bot.dispatch.context import Context
from mubble.bot.dispatch.handler.func import ErrorHandlerT, Func, FuncHandler
from mubble.bot.dispatch.middleware.abc import run_middleware
from mubble.bot.dispatch.middleware.global_middleware import GlobalMiddleware
from mubble.bot.dispatch.view.abc import ABCView
from mubble.bot.dispatch.view.box import (
    CallbackQueryView,
    ChatJoinRequestView,
    ChatMemberView,
    InlineQueryView,
    MessageView,
    PreCheckoutQueryView,
    RawEventView,
    ViewBox,
)
from mubble.modules import logger
from mubble.tools.error_handler.error_handler import ErrorHandler
from mubble.tools.global_context import MubbleContext
from mubble.types.enums import UpdateType
from mubble.types.objects import Update

if typing.TYPE_CHECKING:
    from mubble.bot.cute_types.base import BaseCute
    from mubble.bot.cute_types.update import UpdateCute
    from mubble.bot.rules.abc import ABCRule

T = typing.TypeVar("T", default=typing.Any)
R = typing.TypeVar("R", covariant=True, default=typing.Any)
Event = typing.TypeVar("Event", bound="BaseCute")
P = typing.ParamSpec("P", default=...)

DEFAULT_DATACLASS: typing.Final[type[Update]] = Update


@dataclasses.dataclass(repr=False, kw_only=True)
class Dispatch(
    ABCDispatch,
    ViewBox[
        CallbackQueryView,
        ChatJoinRequestView,
        ChatMemberView,
        InlineQueryView,
        MessageView,
        PreCheckoutQueryView,
        RawEventView,
    ],
    typing.Generic[
        HTTPClient,
        CallbackQueryView,
        ChatJoinRequestView,
        ChatMemberView,
        InlineQueryView,
        MessageView,
        PreCheckoutQueryView,
        RawEventView,
    ],
):
    _global_context: MubbleContext = dataclasses.field(
        init=False,
        default_factory=MubbleContext,
    )
    global_middleware: "GlobalMiddleware" = dataclasses.field(
        default_factory=lambda: GlobalMiddleware(),
    )

    def __repr__(self) -> str:
        return "Dispatch(%s)" % ", ".join(
            f"{k}={v!r}" for k, v in self.get_views().items()
        )

    @property
    def global_context(self) -> MubbleContext:
        return self._global_context

    @property
    def patcher(self) -> Patcher:
        """Alias `patcher` to get `vbml.Patcher` from the global context."""
        return self.global_context.vbml_patcher

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        final: bool = True,
    ) -> typing.Callable[
        [Func[P, R]], FuncHandler["UpdateCute", Func[P, R], ErrorHandler[UpdateCute]]
    ]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        dataclass: type[T],
        final: bool = True,
    ) -> typing.Callable[
        [Func[P, R]], FuncHandler["UpdateCute", Func[P, R], ErrorHandler[T]]
    ]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        update_type: UpdateType,
        final: bool = True,
    ) -> typing.Callable[
        [Func[P, R]], FuncHandler["UpdateCute", Func[P, R], ErrorHandler[UpdateCute]]
    ]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        dataclass: type[T],
        update_type: UpdateType,
        final: bool = True,
    ) -> typing.Callable[
        [Func[P, R]], FuncHandler["UpdateCute", Func[P, R], ErrorHandler[T]]
    ]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        error_handler: ErrorHandlerT,
        final: bool = True,
    ) -> typing.Callable[
        [Func[P, R]], FuncHandler["UpdateCute", Func[P, R], ErrorHandlerT]
    ]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        update_type: UpdateType,
        error_handler: ErrorHandlerT,
        final: bool = True,
    ) -> typing.Callable[
        [Func[P, R]], FuncHandler["UpdateCute", Func[P, R], ErrorHandlerT]
    ]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        dataclass: type[T],
        error_handler: ErrorHandlerT,
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler[T, Func[P, R], ErrorHandlerT]]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        dataclass: type[T],
        update_type: UpdateType,
        error_handler: ErrorHandlerT,
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler[T, Func[P, R], ErrorHandlerT]]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        update_type: UpdateType | None = None,
        dataclass: type[T] = DEFAULT_DATACLASS,
        error_handler: typing.Literal[None] = None,
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler[T, Func[P, R], ErrorHandler[T]]]: ...

    def handle(
        self,
        *rules: "ABCRule",
        update_type: UpdateType | None = None,
        dataclass: type[typing.Any] = DEFAULT_DATACLASS,
        error_handler: ErrorHandlerT | None = None,
        final: bool = True,
    ) -> typing.Callable[..., typing.Any]:
        def wrapper(func):
            handler = FuncHandler(
                func,
                list(rules),
                final=final,
                dataclass=dataclass,
                error_handler=error_handler or ErrorHandler(),
                update_type=update_type,
            )
            self.raw_event.handlers.append(handler)
            return handler

        return wrapper

    async def feed(self, event: Update, api: API[HTTPClient]) -> bool:
        logger.debug(
            "Processing update (update_id={}, update_type={!r})",
            event.update_id,
            event.update_type.name,
        )
        context = Context(raw_update=event)

        if (
            await run_middleware(
                self.global_middleware.pre,
                api,
                event,  # type: ignore
                raw_event=event,
                ctx=context,
                adapter=self.global_middleware.adapter,
            )
            is False
        ):
            return False

        for view in self.get_views().values():
            if await view.check(event):
                logger.debug(
                    "Update (update_id={}, update_type={!r}) matched view {!r}.",
                    event.update_id,
                    event.update_type.name,
                    view,
                )
                if await view.process(event, api, context):
                    return True

        await run_middleware(
            self.global_middleware.post,
            api,
            event,
            raw_event=event,
            ctx=context,
            adapter=self.global_middleware.adapter,
        )

        return False

    def load(self, external: typing.Self) -> None:
        views_external = external.get_views()

        for name, view in self.get_views().items():
            assert (
                name in views_external
            ), f"View {name!r} is undefined in external dispatch."
            view.load(views_external[name])
            setattr(external, name, view)

        self.global_middleware.filters.difference_update(
            external.global_middleware.filters
        )

    def get_view(self, of_type: type[T]) -> Option[T]:
        for view in self.get_views().values():
            if isinstance(view, of_type):
                return Some(view)
        return Nothing()

    def get_views(self) -> dict[str, ABCView]:
        """Get all views."""
        return {
            name: view
            for name, view in self.__dict__.items()
            if isinstance(view, ABCView)
        }

    __call__ = handle


__all__ = ("Dispatch",)



import typing

from mubble.types.objects import Update

from .abc import ABCRule

if typing.TYPE_CHECKING:
    from mubble.tools.adapter.abc import ABCAdapter


class IdRule[Identifier](ABCRule[Identifier]):
    def __init__(
        self,
        adapter: "ABCAdapter[Update, Identifier]",
        tracked_identifiers: set[Identifier] | None = None,
    ):
        self.tracked_identifiers = tracked_identifiers or set()
        self.adapter = adapter

    async def check(self, event: Identifier) -> bool:
        return event in self.tracked_identifiers


__all__ = ("IdRule",)



import dataclasses
import typing

from mubble.types.enums import ProgrammingLanguage

SpecialFormat: typing.TypeAlias = typing.Union[
    "BlockQuote",
    "Link",
    "Mention",
    "PreCode",
    "TgEmoji",
]


def is_spec_format(obj: typing.Any) -> typing.TypeGuard[SpecialFormat]:
    return dataclasses.is_dataclass(obj) and hasattr(obj, "__formatter_name__") and isinstance(obj, Base)


@dataclasses.dataclass(repr=False)
class Base:
    __formatter_name__: typing.ClassVar[str] = dataclasses.field(init=False, repr=False)

    def __repr__(self) -> str:
        return f"<Special formatter {self.__class__.__name__!r} -> {self.__formatter_name__!r}>"


@dataclasses.dataclass(repr=False, slots=True)
class Mention(Base):
    __formatter_name__ = "mention"

    string: str
    user_id: int


@dataclasses.dataclass(repr=False, slots=True)
class Link(Base):
    __formatter_name__ = "link"

    href: str
    string: str | None = None


@dataclasses.dataclass(repr=False, slots=True)
class PreCode(Base):
    __formatter_name__ = "pre_code"

    string: str
    lang: str | ProgrammingLanguage | None = None


@dataclasses.dataclass(repr=False, slots=True)
class TgEmoji(Base):
    __formatter_name__ = "tg_emoji"

    string: str
    emoji_id: int


@dataclasses.dataclass(repr=False, slots=True)
class BlockQuote(Base):
    __formatter_name__ = "block_quote"

    string: str
    expandable: bool = False


__all__ = (
    "Base",
    "BlockQuote",
    "Link",
    "Mention",
    "PreCode",
    "SpecialFormat",
    "TgEmoji",
)



import ssl
import typing

import aiohttp
import certifi
from aiohttp import ClientSession, TCPConnector

import mubble.msgspec_json as json
from mubble.client.abc import ABCClient

if typing.TYPE_CHECKING:
    from aiohttp import ClientResponse

type Data = dict[str, typing.Any] | aiohttp.formdata.FormData
type Response = ClientResponse


class AiohttpClient(ABCClient[aiohttp.formdata.FormData]):
    """HTTP client based on `aiohttp` module."""

    CONNECTION_TIMEOUT_ERRORS = (
        aiohttp.client.ServerConnectionError,
        TimeoutError,
    )
    CLIENT_CONNECTION_ERRORS = (
        aiohttp.client.ClientConnectionError,
        aiohttp.client.ClientConnectorError,
        aiohttp.ClientOSError,
    )

    def __init__(
        self,
        session: ClientSession | None = None,
        timeout: aiohttp.ClientTimeout | None = None,
        **session_params: typing.Any,
    ) -> None:
        self.session = session
        self.session_params = session_params
        self.timeout = timeout or aiohttp.ClientTimeout(total=0)

    def __repr__(self) -> str:
        return "<{}: session={!r}, timeout={}, closed={}>".format(
            self.__class__.__name__,
            self.session,
            self.timeout,
            True if self.session is None else self.session.closed,
        )

    async def request_raw(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        **kwargs: typing.Any,
    ) -> Response:
        if not self.session:
            self.session = ClientSession(
                connector=TCPConnector(ssl=ssl.create_default_context(cafile=certifi.where())),
                json_serialize=json.dumps,
                **self.session_params,
            )

        async with self.session.request(
            url=url,
            method=method,
            data=data,
            timeout=self.timeout,
            **kwargs,
        ) as response:
            await response.read()
            return response

    async def request_json(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        **kwargs: typing.Any,
    ) -> dict[str, typing.Any]:
        response = await self.request_raw(url, method, data, **kwargs)
        return await response.json(
            encoding="UTF-8",
            loads=json.loads,
            content_type=None,
        )

    async def request_text(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        **kwargs: typing.Any,
    ) -> str:
        response = await self.request_raw(url, method, data, **kwargs)  # type: ignore
        return await response.text(encoding="UTF-8")

    async def request_bytes(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        **kwargs: typing.Any,
    ) -> bytes:
        response = await self.request_raw(url, method, data, **kwargs)  # type: ignore
        if response._body is None:
            await response.read()
        return response._body or bytes()

    async def request_content(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        **kwargs: typing.Any,
    ) -> bytes:
        response = await self.request_raw(url, method, data, **kwargs)
        return response._body or bytes()

    async def close(self) -> None:
        if self.session and not self.session.closed:
            await self.session.close()

    @classmethod
    def multipart_form_factory(cls) -> aiohttp.formdata.FormData:
        return aiohttp.formdata.FormData(quote_fields=False)

    def __del__(self) -> None:
        if self.session and not self.session.closed:
            if self.session._connector is not None and self.session._connector_owner:
                self.session._connector._close()
            self.session._connector = None


__all__ = ("AiohttpClient",)



from __future__ import annotations

import asyncio
import dataclasses
import enum
import inspect
from collections import OrderedDict
from functools import wraps

import typing_extensions as typing
from fntypes import Result

from mubble.model import get_params

if typing.TYPE_CHECKING:
    from mubble.bot.rules.abc import ABCRule
    from mubble.node.polymorphic import Polymorphic

type Impl = type[classmethod]
type Function = typing.Callable[..., typing.Any]
type Executor[T] = typing.Callable[
    [T, str, dict[str, typing.Any]],
    typing.Awaitable[Result[typing.Any, typing.Any]],
]

TRANSLATIONS_KEY: typing.Final[str] = "_translations"
MORPH_IMPLEMENTATIONS_KEY = "__morph_implementations__"
IMPL_MARK: typing.Final[str] = "_is_impl"
CONTEXT_KEYS: typing.Final[tuple[str, ...]] = ("ctx", "context")


class FuncParams(typing.TypedDict, total=True):
    args: list[tuple[str, typing.Any | inspect.Parameter.empty]]
    kwargs: list[tuple[str, typing.Any | inspect.Parameter.empty]]
    var_args: typing.NotRequired[str]
    var_kwargs: typing.NotRequired[str]


@dataclasses.dataclass(slots=True, frozen=True)
class Shortcut[T]:
    method_name: str
    executor: Executor[T] | None = dataclasses.field(default=None, kw_only=True)
    custom_params: set[str] = dataclasses.field(default_factory=lambda: set[str](), kw_only=True)


def cache_magic_value(mark_key: str, /):
    def inner[Func: typing.Callable[..., typing.Any]](func: Func) -> Func:
        @wraps(func)
        def wrapper(
            f: Function | None = None,
            /,
            *args: typing.Any,
            **kwargs: typing.Any,
        ) -> typing.Any:
            if f is not None and mark_key not in f.__dict__:
                f.__dict__[mark_key] = func(f, *args, **kwargs)
            return f.__dict__[mark_key]

        return wrapper  # type: ignore

    return inner


def resolve_arg_names(func: Function, start_idx: int = 1) -> tuple[str, ...]:
    return func.__code__.co_varnames[start_idx : func.__code__.co_argcount]


@cache_magic_value("__default_arguments__")
def get_default_args(func: Function, /) -> dict[str, typing.Any]:
    defaults = func.__defaults__
    kwdefaults = {} if not func.__kwdefaults__ else func.__kwdefaults__.copy()
    if not defaults:
        return kwdefaults
    return {
        k: defaults[i] for i, k in enumerate(resolve_arg_names(func, start_idx=0)[-len(defaults) :])
    } | kwdefaults


@cache_magic_value("__function_parameters__")
def get_func_parameters(func: Function, /) -> FuncParams:
    func_params: FuncParams = {"args": [], "kwargs": []}

    for k, p in inspect.signature(func).parameters.items():
        if k in ("self", "cls"):
            continue

        match p.kind:
            case p.POSITIONAL_OR_KEYWORD | p.POSITIONAL_ONLY:
                func_params["args"].append((k, p.default))
            case p.KEYWORD_ONLY:
                func_params["kwargs"].append((k, p.default))
            case p.VAR_POSITIONAL:
                func_params["var_args"] = k
            case p.VAR_KEYWORD:
                func_params["var_kwargs"] = k

    return func_params


def get_annotations(func: Function, *, return_type: bool = False) -> dict[str, typing.Any]:
    annotations = func.__annotations__.copy()
    if not return_type:
        annotations.pop("return", None)
    return annotations


def to_str(s: str | enum.Enum) -> str:
    if isinstance(s, enum.Enum):
        return str(s.value)
    return s


@typing.overload
def magic_bundle(
    function: Function,
    kw: dict[str, typing.Any],
    *,
    start_idx: int = 1,
    omit_defaults: bool = False,
) -> dict[str, typing.Any]: ...


@typing.overload
def magic_bundle(
    function: Function,
    kw: dict[enum.Enum, typing.Any],
    *,
    start_idx: int = 1,
    omit_defaults: bool = False,
) -> dict[str, typing.Any]: ...


@typing.overload
def magic_bundle(
    function: Function,
    kw: dict[str, typing.Any],
    *,
    start_idx: int = 1,
    bundle_ctx: bool = False,
    omit_defaults: bool = False,
) -> dict[str, typing.Any]: ...


@typing.overload
def magic_bundle(
    function: Function,
    kw: dict[enum.Enum, typing.Any],
    *,
    start_idx: int = 1,
    bundle_ctx: bool = False,
    omit_defaults: bool = False,
) -> dict[str, typing.Any]: ...


@typing.overload
def magic_bundle(
    function: Function,
    kw: dict[type[typing.Any], typing.Any],
    *,
    start_idx: int = 1,
    typebundle: typing.Literal[True] = True,
    omit_defaults: bool = False,
) -> dict[str, typing.Any]: ...


def magic_bundle(
    function: Function,
    kw: dict[typing.Any, typing.Any],
    *,
    start_idx: int = 1,
    bundle_ctx: bool = False,
    typebundle: bool = False,
    omit_defaults: bool = False,
) -> dict[str, typing.Any]:
    # Bundle considering the function annotations
    if typebundle:
        return {name: kw[t] for name, t in get_annotations(function, return_type=False).items() if t in kw}

    names = resolve_arg_names(function, start_idx=start_idx)
    # Determine if the function is only have the **kwargs parameter, then bundle all kw
    if "var_kwargs" in get_func_parameters(function) and not names:
        return {to_str(k): v for k, v in kw.items()}

    # Bundle considering the function parameters and defaults (if do not omit defaults)
    defaults = {} if omit_defaults else get_default_args(function)
    args = defaults | {n: v for k, v in kw.items() if (n := to_str(k)) in names}

    # Bundle a context if context key specified in `names`
    if bundle_ctx:
        for key in CONTEXT_KEYS:
            if key in names:
                args[key] = kw
                break

    return args


def join_dicts[Key: typing.Hashable, Value](
    left_dict: dict[Key, typing.Any],
    right_dict: dict[typing.Any, Value],
) -> dict[Key, Value]:
    return {key: right_dict[type_key] for key, type_key in left_dict.items()}


def get_cached_translation[Rule: ABCRule](rule: Rule, locale: str) -> Rule | None:
    return getattr(rule, TRANSLATIONS_KEY, {}).get(locale)


def cache_translation[Rule: ABCRule](
    base_rule: Rule,
    locale: str,
    translated_rule: Rule,
) -> None:
    translations = getattr(base_rule, TRANSLATIONS_KEY, {})
    translations[locale] = translated_rule
    setattr(base_rule, TRANSLATIONS_KEY, translations)


@typing.cast(typing.Callable[..., Impl], lambda f: f)
def impl(method: typing.Callable[..., typing.Any]):
    setattr(method, IMPL_MARK, True)
    return classmethod(method)


def get_polymorphic_implementations(cls: type[Polymorphic], /) -> list[typing.Callable[..., typing.Any]]:
    moprh_impls = getattr(cls, MORPH_IMPLEMENTATIONS_KEY, None)
    if moprh_impls is not None:
        return moprh_impls

    impls = []
    for cls_ in cls.mro():
        impls += [
            obj.__func__
            for obj in vars(cls_).values()
            if isinstance(obj, classmethod) and getattr(obj.__func__, IMPL_MARK, False)
        ]

    setattr(cls, MORPH_IMPLEMENTATIONS_KEY, impls)
    return impls


def shortcut[T](
    method_name: str,
    *,
    executor: Executor[T] | None = None,
    custom_params: set[str] | None = None,
):
    """Decorate a cute method as a shortcut."""

    def wrapper[F: typing.Callable[..., typing.Any]](func: F) -> F:
        @wraps(func)
        async def inner(
            self: T,
            *args: typing.Any,
            **kwargs: typing.Any,
        ) -> typing.Any:
            if executor is None:
                return await func(self, *args, **kwargs)

            params: dict[str, typing.Any] = OrderedDict()
            func_params = get_func_parameters(func)

            for index, (arg, default) in enumerate(func_params["args"]):
                if len(args) > index:
                    params[arg] = args[index]
                elif default is not inspect.Parameter.empty:
                    params[arg] = default

            if var_args := func_params.get("var_args"):
                params[var_args] = args[len(func_params["args"]) :]

            for kwarg, default in func_params["kwargs"]:
                params[kwarg] = (
                    kwargs.pop(kwarg, default) if default is not inspect.Parameter.empty else kwargs.pop(kwarg)
                )

            if var_kwargs := func_params.get("var_kwargs"):
                params[var_kwargs] = kwargs.copy()

            return await executor(self, method_name, get_params(params))

        inner.__shortcut__ = Shortcut(  # type: ignore
            method_name=method_name,
            executor=executor,
            custom_params=custom_params or set(),
        )
        return inner  # type: ignore

    return wrapper


# Source code: https://github.com/facebookincubator/later/blob/main/later/task.py#L75
async def cancel_future(fut: asyncio.Future[typing.Any], /) -> None:
    if fut.done():
        return

    fut.cancel()
    exc: asyncio.CancelledError | None = None

    while not fut.done():
        shielded = asyncio.shield(fut)
        try:
            await asyncio.wait([shielded])
        except asyncio.CancelledError as ex:
            exc = ex
        finally:
            # Insure we handle the exception/value that may exist on the shielded task
            # This will prevent errors logged to the asyncio logger
            if shielded.done() and not shielded.cancelled() and not shielded.exception():
                shielded.result()

    if fut.cancelled():
        if exc is None:
            return
        raise exc from None

    ex = fut.exception()
    if ex is not None:
        raise ex from None

    raise asyncio.InvalidStateError(
        f"Task did not raise CancelledError on cancel: {fut!r} had result {fut.result()!r}",
    )


__all__ = (
    "Shortcut",
    "TRANSLATIONS_KEY",
    "cache_magic_value",
    "cache_translation",
    "cancel_future",
    "get_annotations",
    "get_cached_translation",
    "get_default_args",
    "get_default_args",
    "get_func_parameters",
    "get_polymorphic_implementations",
    "impl",
    "join_dicts",
    "magic_bundle",
    "resolve_arg_names",
    "shortcut",
    "to_str",
)



import inspect
import typing
from contextlib import contextmanager

from mubble.api import API
from mubble.bot.cute_types.update import UpdateCute
from mubble.bot.dispatch.context import Context
from mubble.bot.dispatch.middleware.abc import ABCMiddleware
from mubble.bot.rules.abc import ABCRule, check_rule
from mubble.node import IsNode, compose_nodes
from mubble.tools.adapter.abc import ABCAdapter
from mubble.tools.adapter.raw_update import RawUpdateAdapter
from mubble.types import Update


class GlobalMiddleware(ABCMiddleware):
    adapter = RawUpdateAdapter()

    def __init__(self):
        self.filters: set[ABCRule] = set()
        self.source_filters: dict[ABCAdapter | IsNode, dict[typing.Any, ABCRule]] = {}

    async def pre(self, event: UpdateCute, ctx: Context) -> bool:
        for filter in self.filters:
            if not await check_rule(event.api, filter, event, ctx):
                return False

        # Simple implication.... Grouped by source categories
        for source, identifiers in self.source_filters.items():
            if isinstance(source, ABCAdapter):
                result = source.adapt(event.api, event, ctx)
                if inspect.isawaitable(result):
                    result = await result

                result = result.unwrap_or_none()
                if result is None:
                    return True

            else:
                result = await compose_nodes({"value": source}, ctx, {Update: event, API: event.api})
                if result := result.unwrap():
                    result = result.values["value"]
                else:
                    return True

            if result in identifiers:
                return await check_rule(event.api, identifiers[result], event, ctx)

        return True

    @contextmanager
    def apply_filters(
        self,
        *filters: ABCRule,
        source_filter: tuple[ABCAdapter | IsNode, typing.Any, ABCRule] | None = None,
    ):
        if source_filter is not None:
            self.source_filters.setdefault(source_filter[0], {})
            self.source_filters[source_filter[0]].update({source_filter[1]: source_filter[2]})

        self.filters |= set(filters)
        yield
        self.filters.difference_update(filters)

        if source_filter is not None:  # noqa: SIM102
            if identifiers := self.source_filters.get(source_filter[0]):
                identifiers.pop(source_filter[1], None)


__all__ = ("GlobalMiddleware",)



import typing

from mubble.api.api import API
from mubble.bot.cute_types.base import BaseCute
from mubble.bot.cute_types.update import UpdateCute
from mubble.bot.dispatch.context import Context
from mubble.bot.dispatch.handler.func import Func, FuncHandler
from mubble.bot.dispatch.process import process_inner
from mubble.bot.dispatch.view.abc import ABCEventRawView
from mubble.bot.dispatch.view.base import BaseView
from mubble.bot.rules.abc import ABCRule
from mubble.tools.error_handler.error_handler import ABCErrorHandler, ErrorHandler
from mubble.types.enums import UpdateType
from mubble.types.objects import Update


class RawEventView(ABCEventRawView[UpdateCute], BaseView[UpdateCute]):
    @typing.overload
    def __call__[**P, R](
        self,
        *rules: ABCRule,
        update_type: UpdateType,
        final: bool = True,
    ) -> typing.Callable[
        [Func[P, R]],
        FuncHandler[BaseCute[typing.Any], Func[P, R], ErrorHandler[BaseCute[typing.Any]]],
    ]: ...

    @typing.overload
    def __call__[**P, Dataclass, R](
        self,
        *rules: ABCRule,
        dataclass: type[Dataclass],
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler[Dataclass, Func[P, R], ErrorHandler[Dataclass]]]: ...

    @typing.overload
    def __call__[**P, Dataclass, R](
        self,
        *rules: ABCRule,
        update_type: UpdateType,
        dataclass: type[Dataclass],
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler[Dataclass, Func[P, R], ErrorHandler[Dataclass]]]: ...

    @typing.overload
    def __call__[**P, ErrorHandlerT: ABCErrorHandler, R](
        self,
        *rules: ABCRule,
        error_handler: ErrorHandlerT,
        final: bool = True,
    ) -> typing.Callable[
        [Func[P, R]],
        FuncHandler[BaseCute[typing.Any], Func[P, R], ErrorHandlerT],
    ]: ...

    @typing.overload
    def __call__[**P, ErrorHandlerT: ABCErrorHandler, R](
        self,
        *rules: ABCRule,
        error_handler: ErrorHandlerT,
        update_type: UpdateType,
        final: bool = True,
    ) -> typing.Callable[
        [Func[P, R]],
        FuncHandler[BaseCute[typing.Any], Func[P, R], ErrorHandlerT],
    ]: ...

    @typing.overload
    def __call__[**P, Dataclass, ErrorHandlerT: ABCErrorHandler, R](
        self,
        *rules: ABCRule,
        update_type: UpdateType,
        dataclass: type[Dataclass],
        error_handler: ErrorHandlerT,
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler[Dataclass, Func[P, R], ErrorHandlerT]]: ...

    def __call__(
        self,
        *rules: ABCRule,
        update_type: UpdateType | None = None,
        dataclass: type[typing.Any] | None = None,
        error_handler: ABCErrorHandler | None = None,
        final: bool = True,
    ) -> typing.Callable[..., typing.Any]:
        def wrapper(func: typing.Callable[..., typing.Any]):
            func_handler = FuncHandler(
                func,
                rules=[*self.auto_rules, *rules],
                final=final,
                dataclass=dataclass,
                error_handler=error_handler or ErrorHandler(),
                update_type=update_type,
            )
            self.handlers.append(func_handler)
            return func_handler

        return wrapper

    async def check(self, event: Update) -> bool:
        return bool(self.handlers) or bool(self.middlewares)

    async def process(self, event: Update, api: API, context: Context) -> bool:
        return await process_inner(
            api,
            UpdateCute.from_update(event, bound_api=api),
            event,
            context,
            self.middlewares,
            self.handlers,
            self.return_manager,
        )


__all__ = ("RawEventView",)



import typing
from abc import ABC, abstractmethod

from mubble.api import API
from mubble.bot.dispatch.context import Context
from mubble.tools.adapter.abc import ABCAdapter
from mubble.types.objects import Update


class ABCHandler[Event](ABC):
    final: bool
    adapter: ABCAdapter[Update, Event] | None = None

    @abstractmethod
    async def check(self, api: API, event: Update, ctx: Context | None = None) -> bool:
        pass

    @abstractmethod
    async def run(self, api: API, event: Event, ctx: Context) -> typing.Any:
        pass


__all__ = ("ABCHandler",)



from mubble_old import Token, API, Mubble
from mubble_old.bot.dispatch.handler import MessageReplyHandler
from mubble_old.modules import logger
from mubble_old.rules import Text

api = API(token=Token.from_env())
bot = Mubble(api)
logger.set_level("INFO")

phrases = [
    ("good morning", "Good morning!"),
    ("good bye", "Good bye!"),
    ("hello", "Hi!"),
    ("hi", "Hello!"),
]

bot.on.message.handlers.extend(
    [
        MessageReplyHandler(answer, Text(text, ignore_case=True))
        for text, answer in phrases
    ]
)


bot.run_forever()



import dataclasses
import typing

import msgspec

from mubble.msgspec_utils import encoder
from mubble.types.objects import (
    CallbackGame,
    CopyTextButton,
    KeyboardButtonPollType,
    KeyboardButtonRequestChat,
    KeyboardButtonRequestUsers,
    LoginUrl,
    SwitchInlineQueryChosenChat,
    WebAppInfo,
)

from .callback_data_serilization import ABCDataSerializer, JSONSerializer

if typing.TYPE_CHECKING:
    from _typeshed import DataclassInstance

type CallbackData = str | bytes | dict[str, typing.Any] | DataclassInstance | msgspec.Struct


@dataclasses.dataclass
class BaseButton:
    def get_data(self) -> dict[str, typing.Any]:
        return {k: v for k, v in dataclasses.asdict(self).items() if v is not None}


class RowButtons[KeyboardButton: BaseButton]:
    buttons: list[KeyboardButton]
    auto_row: bool

    def __init__(self, *buttons: KeyboardButton, auto_row: bool = True) -> None:
        self.buttons = list(buttons)
        self.auto_row = auto_row

    def get_data(self) -> list[dict[str, typing.Any]]:
        return [b.get_data() for b in self.buttons]


@dataclasses.dataclass
class Button(BaseButton):
    text: str
    request_contact: bool = dataclasses.field(default=False, kw_only=True)
    request_location: bool = dataclasses.field(default=False, kw_only=True)
    request_chat: KeyboardButtonRequestChat | None = dataclasses.field(
        default=None,
        kw_only=True,
    )
    request_user: KeyboardButtonRequestUsers | None = dataclasses.field(default=None, kw_only=True)
    request_poll: KeyboardButtonPollType | None = dataclasses.field(
        default=None,
        kw_only=True,
    )
    web_app: WebAppInfo | None = dataclasses.field(default=None, kw_only=True)


@dataclasses.dataclass
class InlineButton(BaseButton):
    text: str
    url: str | None = dataclasses.field(default=None, kw_only=True)
    login_url: LoginUrl | None = dataclasses.field(default=None, kw_only=True)
    pay: bool | None = dataclasses.field(default=None, kw_only=True)
    callback_data: CallbackData | None = dataclasses.field(default=None, kw_only=True)
    callback_data_serializer: dataclasses.InitVar[ABCDataSerializer[typing.Any] | None] = dataclasses.field(
        default=None,
        kw_only=True,
    )
    callback_game: CallbackGame | None = dataclasses.field(default=None, kw_only=True)
    copy_text: str | CopyTextButton | None = dataclasses.field(default=None, kw_only=True)
    switch_inline_query: str | None = dataclasses.field(default=None, kw_only=True)
    switch_inline_query_current_chat: str | None = dataclasses.field(default=None, kw_only=True)
    switch_inline_query_chosen_chat: SwitchInlineQueryChosenChat | None = dataclasses.field(
        default=None,
        kw_only=True,
    )
    web_app: str | WebAppInfo | None = dataclasses.field(default=None, kw_only=True)

    def __post_init__(self, callback_data_serializer: ABCDataSerializer[typing.Any] | None) -> None:
        if (
            callback_data_serializer is None
            and isinstance(self.callback_data, msgspec.Struct | dict)
            or dataclasses.is_dataclass(self.callback_data)
        ):
            callback_data_serializer = callback_data_serializer or JSONSerializer(
                self.callback_data.__class__,
            )

        if callback_data_serializer is not None:
            self.callback_data = callback_data_serializer.serialize(self.callback_data)
        elif self.callback_data is not None and not isinstance(self.callback_data, str | bytes):
            self.callback_data = encoder.encode(self.callback_data)

        if isinstance(self.copy_text, str):
            self.copy_text = CopyTextButton(text=self.copy_text)

        if isinstance(self.web_app, str):
            self.web_app = WebAppInfo(url=self.web_app)


__all__ = (
    "BaseButton",
    "Button",
    "InlineButton",
    "RowButtons",
)



import typing
from abc import ABC, abstractmethod

from mubble.api.api import API
from mubble.bot.cute_types.base import BaseCute
from mubble.bot.dispatch.context import Context
from mubble.bot.dispatch.handler.abc import ABCHandler
from mubble.types.objects import Update


class ABCView(ABC):
    def __repr__(self) -> str:
        return "<{}>".format(self.__class__.__name__)

    @abstractmethod
    async def check(self, event: Update) -> bool:
        pass

    @abstractmethod
    async def process(
        self,
        event: Update,
        api: API[typing.Any],
        context: Context,
    ) -> bool:
        pass

    @abstractmethod
    def load(self, external: typing.Self, /) -> None:
        pass


class ABCEventRawView[Event: BaseCute](ABCView, ABC):
    handlers: list[ABCHandler[Event]]


class ABCStateView[Event: BaseCute](ABCView):
    @abstractmethod
    def get_state_key(self, event: Event) -> int | None:
        pass


__all__ = (
    "ABCEventRawView",
    "ABCStateView",
    "ABCView",
)



import dataclasses
import typing

import msgspec

from mubble.api.api import API
from mubble.bot.cute_types import BaseCute
from mubble.msgspec_utils import decoder
from mubble.node.base import ComposeError, FactoryNode
from mubble.types.objects import Update

if typing.TYPE_CHECKING:
    from _typeshed import DataclassInstance

type DataclassType = DataclassInstance | msgspec.Struct | dict[str, typing.Any]


class _EventNode(FactoryNode):
    dataclass: type[DataclassType]
    orig_dataclass: type[DataclassType]

    def __class_getitem__(cls, dataclass: type[DataclassType], /) -> typing.Self:
        return cls(dataclass=dataclass, orig_dataclass=typing.get_origin(dataclass) or dataclass)

    @classmethod
    def compose(cls, raw_update: Update, api: API) -> DataclassType:
        try:
            if issubclass(cls.orig_dataclass, BaseCute):
                update = raw_update if issubclass(cls.orig_dataclass, Update) else raw_update.incoming_update
                dataclass = cls.orig_dataclass.from_update(update=update, bound_api=api)

            elif issubclass(cls.orig_dataclass, msgspec.Struct) or dataclasses.is_dataclass(
                cls.orig_dataclass,
            ):
                dataclass = decoder.convert(
                    obj=raw_update.incoming_update,
                    type=cls.dataclass,
                    from_attributes=True,
                )
            else:
                dataclass = cls.dataclass(**raw_update.incoming_update.to_full_dict())

            return dataclass
        except Exception as exc:
            raise ComposeError(f"Cannot parse an update object into {cls.dataclass!r}, error: {str(exc)}")


if typing.TYPE_CHECKING:
    type EventNode[Dataclass: DataclassType] = Dataclass
else:
    EventNode = _EventNode


__all__ = ("EventNode",)



import typing

from mubble.bot.dispatch.context import Context
from mubble.node.base import IsNode, NodeType, is_node
from mubble.tools.adapter.node import NodeAdapter

from .abc import ABCRule


class NodeRule(ABCRule):
    def __init__(self, *nodes: IsNode | tuple[str, IsNode]) -> None:
        self.nodes: list[IsNode] = []
        self.node_keys: list[str | None] = []

        for binding in nodes:
            node_key, node_t = binding if isinstance(binding, tuple) else (None, binding)
            if not is_node(node_t):
                continue
            self.nodes.append(node_t)
            self.node_keys.append(node_key)

    @property
    def adapter(self) -> NodeAdapter:
        return NodeAdapter(*self.nodes)

    def check(self, resolved_nodes: tuple[NodeType, ...], ctx: Context) -> typing.Literal[True]:
        for i, node in enumerate(resolved_nodes):
            if key := self.node_keys[i]:
                ctx[key] = node
        return True


__all__ = ("NodeRule",)



from mubble.bot.cute_types.pre_checkout_query import PreCheckoutQueryCute
from mubble.bot.dispatch.return_manager.pre_checkout_query import PreCheckoutQueryManager
from mubble.bot.dispatch.view.base import BaseStateView


class PreCheckoutQueryView(BaseStateView[PreCheckoutQueryCute]):
    def __init__(self) -> None:
        super().__init__()
        self.return_manager = PreCheckoutQueryManager()

    @classmethod
    def get_state_key(cls, event: PreCheckoutQueryCute) -> int | None:
        return event.from_user.id


__all__ = ("PreCheckoutQueryView",)



import abc
import typing

from mubble.bot.cute_types.pre_checkout_query import PreCheckoutQueryCute
from mubble.bot.rules.abc import ABCRule, CheckResult
from mubble.tools.adapter.event import EventAdapter
from mubble.types.enums import Currency, UpdateType

PreCheckoutQuery: typing.TypeAlias = PreCheckoutQueryCute


class PaymentInvoiceRule(
    ABCRule[PreCheckoutQuery],
    abc.ABC,
    adapter=EventAdapter(UpdateType.PRE_CHECKOUT_QUERY, PreCheckoutQuery),
):
    @abc.abstractmethod
    def check(self, *args: typing.Any, **kwargs: typing.Any) -> CheckResult: ...


class PaymentInvoiceCurrency(PaymentInvoiceRule):
    def __init__(self, currency: str | Currency, /) -> None:
        self.currency = currency

    def check(self, query: PreCheckoutQuery) -> bool:
        return self.currency == query.currency


__all__ = ("PaymentInvoiceCurrency", "PaymentInvoiceRule")



from mubble.bot.cute_types.callback_query import CallbackQueryCute
from mubble.bot.dispatch.return_manager.callback_query import CallbackQueryReturnManager
from mubble.bot.dispatch.view.base import BaseStateView


class CallbackQueryView(BaseStateView[CallbackQueryCute]):
    def __init__(self) -> None:
        super().__init__()
        self.return_manager = CallbackQueryReturnManager()

    @classmethod
    def get_state_key(cls, event: CallbackQueryCute) -> int | None:
        return event.message_id.unwrap_or_none()


__all__ = ("CallbackQueryView",)
